import pytest

import pikepdf

Dictionary = pikepdf.Dictionary
Name = pikepdf.Name


@pytest.fixture
def graph(resources):
    return pikepdf.open(resources / 'graph.pdf')


@pytest.fixture
def graph_page(graph):
    return pikepdf.Page(graph.pages[0])


def test_page_boxes(graph_page):
    page = graph_page
    assert page.mediabox == page.cropbox == page.trimbox
    page.cropbox = [0, 0, page.mediabox[2] - 100, page.mediabox[3] - 100]
    page.mediabox = [
        page.mediabox[0] - 50,
        page.mediabox[1] - 50,
        page.mediabox[2] + 50,
        page.mediabox[3] + 50,
    ]
    page.trimbox = [50, 50, page.mediabox[2] - 50, page.mediabox[3] - 50]

    assert page.mediabox != page.cropbox
    assert page.cropbox != page.mediabox

    page.cropbox = pikepdf.Array([0, 0, 50, 50])


def test_invalid_boxes(graph_page):
    page = graph_page
    with pytest.raises(ValueError):
        page.mediabox = 'hi'
    with pytest.raises(ValueError):
        page.mediabox = [0, 0, 0]
    with pytest.raises(ValueError):
        page.mediabox = [0, 0, 0, 0, 0, 0]


def test_add_resource(graph_page):
    d = Dictionary(Type=Name.XObject, Subtype=Name.Image, Width=1, Height=1)

    with pytest.raises(ValueError, match="already exists"):
        graph_page.add_resource(d, Name.XObject, Name.Im0, replace_existing=False)

    res = graph_page.add_resource(d, Name.XObject, Name.Im0, replace_existing=True)
    assert graph_page.resources.XObject[res].Width == 1

    res2 = graph_page.add_resource(d, Name.XObject, prefix='Im')
    assert str(res2).startswith("/Im")
    assert graph_page.resources.XObject[res2].Height == 1



import zlib
from io import BytesIO
from os import fspath
from pathlib import Path
from unittest.mock import patch

import pytest
from PIL import Image, ImageCms
from PIL import features as PIL_features

import pikepdf
from pikepdf import (
    Dictionary,
    Name,
    Pdf,
    PdfError,
    PdfImage,
    PdfInlineImage,
    Stream,
    StreamDecodeLevel,
    parse_content_stream,
)
from pikepdf.models.image import (
    DependencyError,
    NotExtractableError,
    UnsupportedImageTypeError,
)

# pylint: disable=redefined-outer-name


def first_image_in(filename):
    pdf = Pdf.open(filename)
    pdfimagexobj = next(iter(pdf.pages[0].images.values()))
    return pdfimagexobj, pdf


@pytest.fixture
def congress(resources):
    return first_image_in(resources / 'congress.pdf')


@pytest.fixture
def sandwich(resources):
    return first_image_in(resources / 'sandwich.pdf')


@pytest.fixture
def jbig2(resources):
    return first_image_in(resources / 'jbig2.pdf')


def test_image_from_nonimage(resources):
    pdf = Pdf.open(resources / 'congress.pdf')
    resources = pdf.pages[0].Contents
    with pytest.raises(TypeError):
        PdfImage(resources)


def test_image(congress):
    pdfimage = PdfImage(congress[0])
    pillowimage = pdfimage.as_pil_image()

    assert pillowimage.mode == pdfimage.mode
    assert pillowimage.size == pdfimage.size


def test_imagemask(congress):
    assert PdfImage(congress[0]).image_mask == False


def test_image_replace(congress, outdir):
    pdfimage = PdfImage(congress[0])
    pillowimage = pdfimage.as_pil_image()

    grayscale = pillowimage.convert('L')
    grayscale = grayscale.resize((4, 4))  # So it is not obnoxious on error

    congress[0].write(zlib.compress(grayscale.tobytes()), filter=Name("/FlateDecode"))
    congress[0].ColorSpace = Name("/DeviceGray")
    pdf = congress[1]
    pdf.save(outdir / 'congress_gray.pdf')


def test_lowlevel_jpeg(congress):
    raw_bytes = congress[0].read_raw_bytes()
    with pytest.raises(PdfError):
        congress[0].read_bytes()

    im = Image.open(BytesIO(raw_bytes))
    assert im.format == 'JPEG'

    pim = PdfImage(congress[0])
    b = BytesIO()
    pim.extract_to(stream=b)
    b.seek(0)
    im = Image.open(b)
    assert im.size == (congress[0].Width, congress[0].Height)
    assert im.mode == 'RGB'


def test_lowlevel_replace_jpeg(congress, outdir):
    # This test will modify the PDF so needs its own image
    raw_bytes = congress[0].read_raw_bytes()

    im = Image.open(BytesIO(raw_bytes))
    grayscale = im.convert('L')
    grayscale = grayscale.resize((4, 4))  # So it is not obnoxious on error

    congress[0].write(
        zlib.compress(grayscale.tobytes()[:10]), filter=Name("/FlateDecode")
    )
    congress[0].ColorSpace = Name('/DeviceGray')

    pdf = congress[1]
    pdf.save(outdir / 'congress_gray.pdf')


@pytest.fixture
def inline(resources):
    pdf = Pdf.open(resources / 'image-mono-inline.pdf')
    for operands, _command in parse_content_stream(pdf.pages[0]):
        if operands and isinstance(operands[0], PdfInlineImage):
            return operands[0], pdf


def test_inline(inline):
    iimage, pdf = inline
    assert iimage.width == 8
    assert iimage.image_mask == False
    assert iimage.mode == 'RGB'
    assert iimage.is_inline
    assert iimage.colorspace == '/DeviceRGB'

    unparsed = iimage.unparse()

    cs = pdf.make_stream(unparsed)
    for operands, _command in parse_content_stream(cs):
        if operands and isinstance(operands[0], PdfInlineImage):
            reparsed_iim = operands[0]
            assert reparsed_iim == iimage


def test_bits_per_component_missing(congress):
    cong_im = congress[0]
    del cong_im.stream_dict['/BitsPerComponent']
    assert PdfImage(congress[0]).bits_per_component == 8


@pytest.mark.parametrize(
    'w,h,pixeldata,cs,bpc',
    [
        (1, 1, b'\xff', '/DeviceGray', 1),
        (1, 1, b'\xf0', '/DeviceGray', 8),
        (1, 1, b'\xff\x00\xff', '/DeviceRGB', 8),
    ],
)
def test_image_roundtrip(outdir, w, h, pixeldata, cs, bpc):
    pdf = Pdf.new()

    image_data = pixeldata * (w * h)

    image = Stream(pdf, image_data)
    image.Type = Name('/XObject')
    image.Subtype = Name('/Image')
    image.ColorSpace = Name(cs)
    image.BitsPerComponent = bpc
    image.Width = w
    image.Height = h

    xobj = {'/Im1': image}
    resources = {'/XObject': xobj}
    mediabox = [0, 0, 100, 100]
    stream = b'q 100 0 0 100 0 0 cm /Im1 Do Q'
    contents = Stream(pdf, stream)

    page_dict = {
        '/Type': Name('/Page'),
        '/MediaBox': mediabox,
        '/Contents': contents,
        '/Resources': resources,
    }
    page = pdf.make_indirect(page_dict)

    pdf.pages.append(page)
    outfile = outdir / f'test{w}{h}{cs[1:]}{bpc}.pdf'
    pdf.save(
        outfile, compress_streams=False, stream_decode_level=StreamDecodeLevel.none
    )

    with Pdf.open(outfile) as p2:
        pim = PdfImage(p2.pages[0].Resources.XObject['/Im1'])

        assert pim.bits_per_component == bpc
        assert pim.colorspace == cs
        assert pim.width == w
        assert pim.height == h
        if cs == '/DeviceRGB':
            assert pim.mode == 'RGB'
        elif cs == '/DeviceGray' and bpc == 8:
            assert pim.mode == 'L'
        elif bpc == 1:
            assert pim.mode == '1'
        assert not pim.palette

        assert pim.filters == []
        assert pim.read_bytes() == pixeldata

        outstream = BytesIO()
        pim.extract_to(stream=outstream)
        outstream.seek(0)
        im = Image.open(outstream)
        assert pim.mode == im.mode


@pytest.mark.parametrize(
    'filename,bpc,filters,ext,mode,format_',
    [
        ('sandwich.pdf', 1, ['/CCITTFaxDecode'], '.tif', '1', 'TIFF'),
        ('congress-gray.pdf', 8, ['/DCTDecode'], '.jpg', 'L', 'JPEG'),
        ('congress.pdf', 8, ['/DCTDecode'], '.jpg', 'RGB', 'JPEG'),
        ('cmyk-jpeg.pdf', 8, ['/DCTDecode'], '.jpg', 'CMYK', 'JPEG'),
    ],
)
def test_direct_extract(resources, filename, bpc, filters, ext, mode, format_):
    xobj, _pdf = first_image_in(resources / filename)
    pim = PdfImage(xobj)

    assert pim.bits_per_component == bpc
    assert pim.filters == filters

    outstream = BytesIO()
    outext = pim.extract_to(stream=outstream)
    assert outext == ext, 'unexpected output file'
    outstream.seek(0)

    im = Image.open(outstream)
    assert im.mode == mode
    assert im.format == format_


@pytest.mark.parametrize(
    'filename,bpc,rgb',
    [
        ('pal.pdf', 8, (0, 0, 255)),
        ('pal-1bit-trivial.pdf', 1, (255, 255, 255)),
        ('pal-1bit-rgb.pdf', 1, (255, 128, 0)),
    ],
)
def test_image_palette(resources, filename, bpc, rgb):
    pdf = Pdf.open(resources / filename)
    pim = PdfImage(next(iter(pdf.pages[0].images.values())))

    assert pim.palette[0] == 'RGB'
    assert pim.colorspace == '/DeviceRGB'
    assert not pim.is_inline
    assert pim.mode == 'P'
    assert pim.bits_per_component == bpc

    outstream = BytesIO()
    pim.extract_to(stream=outstream)

    im = pim.as_pil_image().convert('RGB')
    assert im.getpixel((1, 1)) == rgb


def test_bool_in_inline_image():
    piim = PdfInlineImage(image_data=b'', image_object=(Name.IM, True))
    assert piim.image_mask


@pytest.mark.skipif(
    not PIL_features.check_codec('jpg_2000'), reason='no JPEG2000 codec'
)
def test_jp2(resources):
    pdf = Pdf.open(resources / 'pike-jp2.pdf')
    xobj = next(iter(pdf.pages[0].images.values()))
    pim = PdfImage(xobj)

    assert '/JPXDecode' in pim.filters
    assert pim.colorspace == '/DeviceRGB'
    assert not pim.is_inline
    assert not pim.indexed
    assert pim.mode == 'RGB'
    assert pim.bits_per_component == 8

    outstream = BytesIO()
    pim.extract_to(stream=outstream)
    del pim
    del xobj.ColorSpace

    # If there is no explicit ColorSpace metadata we should get it from the
    # compressed data stream
    pim = PdfImage(xobj)
    assert pim.colorspace == '/DeviceRGB'
    assert pim.bits_per_component == 8


def test_extract_filepath(congress, outdir):
    xobj, _pdf = congress
    pim = PdfImage(xobj)

    # fspath is for Python 3.5
    result = pim.extract_to(fileprefix=fspath(outdir / 'image'))
    assert Path(result).exists()
    assert (outdir / 'image.jpg').exists()


def test_extract_direct_fails_nondefault_colortransform(congress):
    xobj, _pdf = congress

    xobj.DecodeParms = Dictionary(
        ColorTransform=42  # Non standard (or allowed in the spec)
    )
    pim = PdfImage(xobj)

    bio = BytesIO()
    with pytest.raises(NotExtractableError):
        pim._extract_direct(stream=bio)
    with pytest.raises(UnsupportedImageTypeError):
        pim.extract_to(stream=bio)

    xobj.ColorSpace = Name.DeviceCMYK
    pim = PdfImage(xobj)
    with pytest.raises(NotExtractableError):
        pim._extract_direct(stream=bio)
    with pytest.raises(UnsupportedImageTypeError):
        pim.extract_to(stream=bio)


def test_icc_use(resources):
    xobj, _pdf = first_image_in(resources / '1biticc.pdf')

    pim = PdfImage(xobj)
    assert pim.mode == '1'
    assert pim.colorspace == '/ICCBased'
    assert pim.bits_per_component == 1

    assert pim.icc.profile.xcolor_space == 'GRAY'


def test_icc_extract(resources):
    xobj, _pdf = first_image_in(resources / 'aquamarine-cie.pdf')

    pim = PdfImage(xobj)
    assert pim.as_pil_image().info['icc_profile'] == pim.icc.tobytes()


def test_icc_palette(resources):
    xobj, _pdf = first_image_in(resources / 'pink-palette-icc.pdf')
    pim = PdfImage(xobj)
    assert pim.icc.profile.xcolor_space == 'RGB '  # with trailing space
    b = BytesIO()
    pim.extract_to(stream=b)
    b.seek(0)

    im = Image.open(b)
    assert im.size == (xobj.Width, xobj.Height)
    assert im.mode == 'P'
    pil_icc = im.info.get('icc_profile')
    pil_icc_stream = BytesIO(pil_icc)
    pil_prf = ImageCms.ImageCmsProfile(pil_icc_stream)

    assert pil_prf.tobytes() == pim.icc.tobytes()


def test_stacked_compression(resources):
    xobj, _pdf = first_image_in(resources / 'pike-flate-jp2.pdf')

    pim = PdfImage(xobj)
    assert pim.mode == 'RGB'
    assert pim.colorspace == '/DeviceRGB'
    assert pim.bits_per_component == 8
    assert pim.filters == ['/FlateDecode', '/JPXDecode']


def test_ccitt_photometry(sandwich):
    xobj, _pdf = sandwich

    pim = PdfImage(xobj)
    im = pim.as_pil_image()
    im = im.convert('L')
    assert im.getpixel((0, 0)) == 255, "Expected white background"


def test_ccitt_encodedbytealign(sandwich):
    xobj, _pdf = sandwich

    # Pretend this is image is "EncodedByteAlign". We don't have a FOSS
    # example of such an image.
    xobj.DecodeParms.EncodedByteAlign = True
    pim = PdfImage(xobj)
    with pytest.raises(UnsupportedImageTypeError):
        pim.as_pil_image()


def test_imagemagick_uses_rle_compression(resources):
    xobj, _rle = first_image_in(resources / 'rle.pdf')

    pim = PdfImage(xobj)
    im = pim.as_pil_image()
    assert im.getpixel((5, 5)) == (255, 128, 0)


def test_jbig2_not_available(jbig2, monkeypatch):
    xobj, _pdf = jbig2
    pim = PdfImage(xobj)

    monkeypatch.setattr(pikepdf.jbig2, 'jbig2dec_available', lambda: False)
    with pytest.raises(DependencyError):
        pim.as_pil_image()


needs_jbig2dec = pytest.mark.skipif(
    not pikepdf.jbig2.jbig2dec_available(), reason="jbig2dec not installed"
)


@needs_jbig2dec
def test_jbig2(jbig2):
    xobj, _pdf = jbig2
    pim = PdfImage(xobj)
    im = pim.as_pil_image()
    assert im.size == (1000, 1520)
    assert im.getpixel((0, 0)) == 0  # Ensure loaded


@needs_jbig2dec
def test_jbig2_global(resources):
    xobj, _pdf = first_image_in(resources / 'jbig2global.pdf')
    pim = PdfImage(xobj)
    im = pim.as_pil_image()
    assert im.size == (4000, 2864)
    assert im.getpixel((0, 0)) == 255  # Ensure loaded


@needs_jbig2dec
def test_jbig2_global_palette(resources):
    xobj, _pdf = first_image_in(resources / 'jbig2global.pdf')
    xobj.ColorSpace = pikepdf.Array(
        [Name.Indexed, Name.DeviceRGB, 1, b'\x00\x00\x00\xff\xff\xff']
    )
    pim = PdfImage(xobj)
    im = pim.as_pil_image()
    assert im.size == (4000, 2864)
    assert im.getpixel((0, 0)) == 255  # Ensure loaded


def test_ccitt_icc(resources):
    xobj, pdf = first_image_in(resources / 'sandwich.pdf')

    pim = PdfImage(xobj)
    assert pim.icc is None
    bio = BytesIO()
    output_type = pim.extract_to(stream=bio)
    assert output_type == '.tif'
    bio.seek(0)
    assert b'GRAYXYZ' not in bio.read(1000)
    bio.seek(0)
    assert Image.open(bio)

    icc_data = (resources / 'Gray.icc').read_bytes()
    icc_stream = pdf.make_stream(icc_data)
    icc_stream.N = 1
    xobj.ColorSpace = pikepdf.Array([Name.ICCBased, icc_stream])

    pim = PdfImage(xobj)
    assert pim.icc.profile.xcolor_space == 'GRAY'
    bio = BytesIO()
    output_type = pim.extract_to(stream=bio)
    assert output_type == '.tif'
    bio.seek(0)
    assert b'GRAYXYZ' in bio.read(1000)
    bio.seek(0)
    assert Image.open(bio)


def test_invalid_icc(resources):
    xobj, _pdf = first_image_in(resources / 'pink-palette-icc.pdf')

    cs = xobj.ColorSpace[1][1]  # [/Indexed [/ICCBased <stream>]]
    cs.write(b'foobar')  # corrupt the ICC profile
    with pytest.raises(
        UnsupportedImageTypeError, match="ICC profile corrupt or not readable"
    ):
        pim = PdfImage(xobj)
        _icc = pim.icc



import pytest

from pikepdf import Dictionary, ForeignObjectError, Name, Object, Pdf, Stream


@pytest.fixture
def vera(resources):
    # Has XMP but no docinfo
    return Pdf.open(resources / 'veraPDF test suite 6-2-10-t02-pass-a.pdf')


@pytest.fixture
def outlines(resources):
    return Pdf.open(resources / 'outlines.pdf')


def test_no_foreign_on_direct(vera):
    direct_object = Dictionary()
    with pytest.raises(ForeignObjectError, match="called with direct object"):
        vera.copy_foreign(direct_object)


def test_must_use_copy_foreign(vera, outlines, outpdf):
    vera.Root.Names = Dictionary()
    vera.Root.Names.Dests = outlines.Root.Names.Dests
    with pytest.raises(ForeignObjectError, match="add objects from another file"):
        vera.save(outpdf)


def test_self_copy_foreign(vera):
    direct_object = Dictionary()
    indirect_object = vera.make_indirect(direct_object)
    assert indirect_object.is_indirect
    with pytest.raises(ForeignObjectError, match="called with object from"):
        vera.Root.IndirectObj = vera.copy_foreign(indirect_object)


def test_copy_foreign_copies(vera, outlines, outpdf):
    assert outlines.Root.Names.is_indirect
    assert outlines.Root.Names.is_owned_by(outlines)

    vera.Root.Names = vera.copy_foreign(outlines.Root.Names)
    assert vera.Root.Names.is_owned_by(vera)
    assert not outlines.Root.Names.is_owned_by(vera)
    vera.save(outpdf)



import pytest

from pikepdf import Pdf


@pytest.fixture
def vera(resources):
    # A file that is not linearized
    return Pdf.open(resources / 'veraPDF test suite 6-2-10-t02-pass-a.pdf')


def test_foreign_linearization(vera):
    assert not vera.is_linearized
    with pytest.raises(RuntimeError, match="not linearized"):
        vera.check_linearization()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)

"""A library for manipulating PDFs

isort:skip_file
"""

import os

try:
    from . import _qpdf
except ImportError as _e:  # pragma: no cover
    msg = "pikepdf's extension library failed to import"
    raise ImportError(msg) from _e

try:
    from ._version import __version__
except ImportError as _e:  # pragma: no cover
    raise ImportError("Failed to determine version") from _e

from ._qpdf import (
    AccessMode,
    ForeignObjectError,
    ObjectStreamMode,
    Page,
    PasswordError,
    Pdf,
    PdfError,
    StreamDecodeLevel,
    Token,
    TokenFilter,
    TokenType,
)

from .objects import (
    Array,
    Dictionary,
    Name,
    Object,
    ObjectType,
    Operator,
    Stream,
    String,
)

from .models import (
    Encryption,
    Outline,
    OutlineItem,
    OutlineStructureError,
    PageLocation,
    PdfImage,
    PdfInlineImage,
    PdfMatrix,
    Permissions,
    UnsupportedImageTypeError,
    make_page_destination,
    parse_content_stream,
    unparse_content_stream,
)

from . import _methods, codec

__libqpdf_version__ = _qpdf.qpdf_version()


# Provide pikepdf.{open, new} -> pikepdf.Pdf.open
open = Pdf.open  # pylint: disable=redefined-builtin
new = Pdf.new

# Exclude .open, .new here from to make sure from pikepdf import * does not clobber
# builtins.open()
__all__ = [
    'AccessMode',
    'Array',
    'Dictionary',
    'Encryption',
    'Name',
    'Object',
    'ObjectStreamMode',
    'ObjectType',
    'Operator',
    'Outline',
    'OutlineItem',
    'OutlineStructureError',
    'Page',
    'PageLocation',
    'PasswordError',
    'Pdf',
    'PdfError',
    'PdfImage',
    'PdfInlineImage',
    'PdfMatrix',
    'Permissions',
    'Stream',
    'StreamDecodeLevel',
    'String',
    'Token',
    'TokenFilter',
    'TokenType',
    'UnsupportedImageTypeError',
    'make_page_destination',
    'models',
    'parse_content_stream',
    'unparse_content_stream',
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)

"""Provide classes to stand in for PDF objects

The purpose of these is to provide nice-looking classes to allow explicit
construction of PDF objects and more pythonic idioms and facilitate discovery
by documentation generators and linters.

It's also a place to narrow the scope of input types to those more easily
converted to C++.

There is some deliberate "smoke and mirrors" here: all of the objects are truly
instances of ``pikepdf.Object``, which is a variant container object. The
``__new__`` constructs a ``pikepdf.Object`` in each case, and the rest of the
class definition is present as an aide for code introspection.
"""

# pylint: disable=unused-import, abstract-method

from secrets import token_urlsafe
from typing import TYPE_CHECKING, Iterable, Optional, Union
from warnings import warn

from . import _qpdf
from ._qpdf import Object, ObjectType

if TYPE_CHECKING:  # pragma: no cover
    from pikepdf import Pdf

# By default pikepdf.Object will identify itself as pikepdf._qpdf.Object
# Here we change the module to discourage people from using that internal name
# Instead it will become pikepdf.objects.Object
Object.__module__ = __name__
ObjectType.__module__ = __name__


# type(Object) is the metaclass that pybind11 defines; we wish to extend that
# pylint cannot see the C++ metaclass definition is thoroughly confused.
# pylint: disable=invalid-metaclass


class _ObjectMeta(type(Object)):  # type: ignore
    """Supports instance checking"""

    def __instancecheck__(self, instance):
        # Note: since this class is a metaclass, self is a class object
        if type(instance) != Object:
            return False
        return self.object_type == instance._type_code


class _NameObjectMeta(_ObjectMeta):
    """Supports usage pikepdf.Name.Whatever -> Name('/Whatever')"""

    def __getattr__(self, attr):
        if attr.startswith('_'):
            return getattr(_ObjectMeta, attr)
        return Name('/' + attr)

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            return _ObjectMeta.__setattr__(attr, value)
        raise TypeError("Attributes may not be set on pikepdf.Name")

    def __getitem__(self, item):
        if item.startswith('/'):
            item = item[1:]
        raise TypeError(
            "pikepdf.Name is not subscriptable. You probably meant:\n"
            f"    pikepdf.Name.{item}\n"
            "or\n"
            f"    pikepdf.Name('/{item}')\n"
        )


class Name(Object, metaclass=_NameObjectMeta):
    """Constructs a PDF Name object

    Names can be constructed with two notations:

        1. ``Name.Resources``

        2. ``Name('/Resources')``

    The two are semantically equivalent. The former is preferred for names
    that are normally expected to be in a PDF. The latter is preferred for
    dynamic names and attributes.
    """

    object_type = ObjectType.name

    def __new__(cls, name: str):
        # QPDF_Name::unparse ensures that names are always saved in a UTF-8
        # compatible way, so we only need to guard the input.
        if isinstance(name, bytes):
            raise TypeError("Name should be str")
        return _qpdf._new_name(name)

    @classmethod
    def random(cls, len_: int = 16, prefix: str = ''):
        """Generate a cryptographically strong random, valid PDF Name.

        When the length paramater is 16 (16 random bytes or 128 bits), the result
        is probably globally unique and can be treated as never colliding with
        other names.
        """
        # It so happens that urlsafe names are also safe names for PDF Name objects
        return _qpdf._new_name(f"/{prefix}{token_urlsafe(len_)}")


class Operator(Object, metaclass=_ObjectMeta):
    """Constructs an operator for use in a content stream.

    An Operator is one of a limited set of commands that can appear in PDF content
    streams (roughly the mini-language that draws objects, lines and text on a
    virtual PDF canvas). The commands :func:`parse_content_stream` and
    :func:`unparse_content_stream` create and expect Operators respectively, along
    with their operands.

    pikepdf uses the special Operator "INLINE IMAGE" to denote an inline image
    in a content stream.
    """

    object_type = ObjectType.operator

    def __new__(cls, name: str):
        return _qpdf._new_operator(name)


class String(Object, metaclass=_ObjectMeta):
    """Constructs a PDF String object"""

    object_type = ObjectType.string

    def __new__(cls, s: Union[str, bytes]):
        """
        Args:
            s: The string to use. String will be encoded for
                PDF, bytes will be constructed without encoding.

        Returns:
            pikepdf.Object
        """
        if isinstance(s, bytes):
            return _qpdf._new_string(s)
        return _qpdf._new_string_utf8(s)


class Array(Object, metaclass=_ObjectMeta):
    """Constructs a PDF Array object"""

    object_type = ObjectType.array

    def __new__(cls, a: Optional[Iterable] = None):
        """
        Args:
            a: An iterable of objects. All objects must be either
                `pikepdf.Object` or convertible to `pikepdf.Object`.

        Returns:
            pikepdf.Object
        """

        if isinstance(a, (str, bytes)):
            raise TypeError('Strings cannot be converted to arrays of chars')
        if a is None:
            a = []
        return _qpdf._new_array(a)


class Dictionary(Object, metaclass=_ObjectMeta):
    """Constructs a PDF Dictionary object"""

    object_type = ObjectType.dictionary

    def __new__(cls, d=None, **kwargs):
        """
        Constructs a PDF Dictionary from either a Python ``dict`` or keyword
        arguments.

        These two examples are equivalent:

        .. code-block:: python

            pikepdf.Dictionary({'/NameOne': 1, '/NameTwo': 'Two'})

            pikepdf.Dictionary(NameOne=1, NameTwo='Two')

        In either case, the keys must be strings, and the strings
        correspond to the desired Names in the PDF Dictionary. The values
        must all be convertible to `pikepdf.Object`.

        Returns:
            pikepdf.Object
        """
        if kwargs and d is not None:
            raise ValueError('Unsupported parameters')
        if kwargs:
            # Add leading slash
            # Allows Dictionary(MediaBox=(0,0,1,1), Type=Name('/Page')...
            return _qpdf._new_dictionary({('/' + k): v for k, v in kwargs.items()})
        if not d:
            d = {}
        if d and any(not key.startswith('/') for key in d.keys()):
            raise ValueError("Dictionary created from strings must begin with '/'")
        return _qpdf._new_dictionary(d)


class Stream(Object, metaclass=_ObjectMeta):
    """Constructs a PDF Stream object"""

    object_type = ObjectType.stream

    def __new__(cls, owner: 'Pdf', data: bytes = None, d=None, **kwargs):
        """
        Args:
            owner: The Pdf to which this stream shall be attached.
            obj: The data bytes for the stream.
            d: A mapping object that will be used to construct a ``Dictionary``.
            kwargs: Keyword arguments that will define the dictionary. Do not set
                /Filter or /Length here as pikepdf will manage these.
        Returns:
            pikepdf.Object
        """

        # Support __new__(...obj=bytes) which should have been data=bytes,
        # drop in pikepdf 3
        if 'obj' in kwargs:  # pragma: no cover
            warn("Deprecated parameter 'obj', use 'data' instead", DeprecationWarning)
            if data is None:
                data = kwargs['obj']
            del kwargs['obj']

        if data is None:
            raise TypeError("Must make Stream from binary data")

        stream = _qpdf._new_stream(owner, data)
        if d or kwargs:
            stream_dict = Dictionary(d, **kwargs)
            stream.stream_dict = stream_dict
        return stream



import sys
from glob import glob
from os import environ
from os.path import dirname, join

from setuptools import find_packages, setup

try:
    from pybind11.setup_helpers import ParallelCompile, Pybind11Extension, build_ext
except ImportError:
    from setuptools import Extension as Pybind11Extension
    from setuptools.command.build_ext import build_ext

    ParallelCompile = None
    print("pybind11.setup_helpers NOT loaded - you might need to pip install pybind11")

extra_includes = []
extra_library_dirs = []
qpdf_source_tree = environ.get('QPDF_SOURCE_TREE', None)
if qpdf_source_tree:
    # Point this to qpdf source tree built with shared libaries
    extra_includes.append(join(qpdf_source_tree, 'include'))
    extra_library_dirs.append(join(qpdf_source_tree, 'libqpdf/build/.libs'))
if 'bsd' in sys.platform:
    extra_includes.append('/usr/local/include')

try:
    from setuptools_scm import get_version

    __version__ = get_version()
except ImportError:
    __version__ = '0.0.1'


ext_modules = [
    Pybind11Extension(
        'pikepdf._qpdf',
        sorted(glob('src/qpdf/*.cpp')),
        depends=sorted(glob('src/qpdf/*.h')),
        include_dirs=[
            # Path to pybind11 headers
            *extra_includes,
        ],
        library_dirs=[*extra_library_dirs],
        libraries=['qpdf'],
        cxx_std=14,
    )
]

if sys.platform == 'cygwin':
    # On cygwin, use gnu++14 instead of c++14
    eca = ext_modules[0].extra_compile_args
    eca[eca.index('-std=c++14')] = '-std=gnu++14'

# Debug build
# ext_modules[0].extra_compile_args.append('-g3')

setup_py_cwd = dirname(__file__)

with open(join(setup_py_cwd, 'requirements/docs.txt')) as f:
    docs_require = [
        line.strip() for line in f if line.strip() and not line.strip().startswith('#')
    ]


with open(join(setup_py_cwd, 'requirements/test.txt')) as f:
    tests_require = [
        line.strip() for line in f if line.strip() and not line.strip().startswith('#')
    ]

with open(join(setup_py_cwd, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

if __name__ == '__main__':
    if ParallelCompile:
        ParallelCompile().install()
    setup(
        name='pikepdf',
        author='James R. Barlow',
        author_email='james@purplerock.ca',
        url='https://github.com/pikepdf/pikepdf',
        description='Read and write PDFs with Python, powered by qpdf',
        long_description=readme,
        long_description_content_type='text/markdown',
        ext_modules=ext_modules,
        install_requires=[
            'lxml >= 4.0',
            'Pillow >= 6',  # only needed for manipulating images
        ],
        setup_requires=[
            'setuptools >= 50',
            'wheel >= 0.35',
            'setuptools_scm[toml] >= 4.1',
            'setuptools_scm_git_archive',
            'pybind11 >= 2.6.0, <3',
        ],
        extras_require={'docs': docs_require},
        zip_safe=False,
        python_requires='>=3.6',
        use_scm_version=True,
        tests_require=tests_require,
        cmd_class={"build_ext": build_ext},
        package_dir={'': 'src'},
        packages=find_packages('src'),
        package_data={'': ['*.txt'], 'pikepdf': ['*.dll', 'py.typed']},
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Programming Language :: C++",
            "Topic :: Multimedia :: Graphics",
            "Topic :: Software Development :: Libraries",
        ],
        project_urls={
            'Documentation': 'https://pikepdf.readthedocs.io/',
            'Source': 'https://github.com/pikepdf/pikepdf',
            'Tracker': 'https://github.com/pikepdf/pikepdf/issues',
        },
    )



from typing import (
    Tuple,
    Any,
    Collection,
    Iterable,
    Iterator,
    List,
    Callable,
    Set,
    Union,
    TypeVar,
    overload,
)
from enum import Enum

from pikepdf.models.image import PdfInlineImage

T = TypeVar('T', bound=Object)


def _new_operator(op: str) -> Object:
    ...


def _Null() -> Any:
    ...


def _encode(handle: Any) -> Object:
    ...


def _new_array(arg0: Iterable) -> Object:
    ...


def _new_boolean(arg0: bool) -> Object:
    ...


def _new_dictionary(arg0: dict) -> Object:
    ...


def _new_integer(arg0: int) -> Object:
    ...


def _new_name(arg0: str) -> Object:
    ...


@overload
def _new_real(arg0: str) -> Object:
    ...


@overload
def _new_real(value: float, places: int = ...) -> Object:
    ...


@overload
def _new_real(*args, **kwargs) -> Any:
    ...


def _new_stream(arg0: Pdf, arg1: bytes) -> Object:
    ...


def _new_string(s: Union[str, bytes]) -> Object:
    ...


def _new_string_utf8(s: str) -> Object:
    ...


def _roundtrip(obj: Any) -> Any:
    ...


def _test_file_not_found(*args, **kwargs) -> Any:
    ...


def get_decimal_precision() -> int:
    ...


def pdf_doc_to_utf8(pdfdoc: bytes) -> str:
    ...


def qpdf_version() -> str:
    ...


def set_access_default_mmap(mmap: bool) -> bool:
    ...


def set_decimal_precision(prec: int) -> int:
    ...


def unparse(obj: Any) -> bytes:
    ...


def utf8_to_pdf_doc(utf8: str, unknown: bytes) -> Tuple[bool, bytes]:
    ...


class AccessMode(Enum):
    default: int = ...
    mmap: int = ...
    mmap_only: int = ...
    stream: int = ...


class Annotation:
    def __init__(self, arg0: Object) -> None:
        ...

    def get_appearance_stream(self, which: Object, state: str = ...) -> Object:
        ...

    def get_page_content_for_appearance(
        self,
        name: Object,
        rotate: int,
        required_flags: int = ...,
        forbidden_flags: int = ...,
    ) -> str:
        ...

    @property
    def appearance_dict(self) -> Object:
        ...

    @property
    def appearance_state(self) -> str:
        ...

    @property
    def flags(self) -> int:
        ...

    @property
    def subtype(self) -> str:
        ...


class Buffer:
    def __init__(self, *args, **kwargs) -> None:
        ...


class EncryptionMethod(Enum):
    none: int = ...
    unknown: int = ...
    rc4: int = ...
    aes: int = ...
    aesv3: int = ...


class Object:
    _ipython_key_completions_: Any = ...
    _repr_mimebundle_: Any = ...
    emplace: Any = ...
    write: Any = ...

    def __init__(self, *args, **kwargs) -> None:
        ...

    def _inline_image_raw_bytes(self) -> bytes:
        ...

    def _parse_page_contents(self, callbacks: Callable) -> None:
        ...

    def _parse_page_contents_grouped(
        self, whitelist: str
    ) -> List[Tuple[Collection[Union[Object, 'PdfInlineImage']], Operator]]:
        ...

    def _parse_stream(self, *args, **kwargs) -> Any:
        ...

    def _parse_stream_grouped(self, *args, **kwargs) -> Any:
        ...

    def _write(self, data: bytes, filter: object, decode_parms: object) -> None:
        ...

    def append(self, pyitem: Object) -> None:
        ...

    def as_dict(self) -> _ObjectMapping:
        ...

    def as_list(self) -> _ObjectList:
        ...

    def extend(self, arg0: Iterable[Object]) -> None:
        ...

    @overload
    def get(self, key: str, default: object = ...) -> object:
        ...

    @overload
    def get(key, default=...) -> Any:
        ...

    @overload
    def get(self, key: Object, default: object = ...) -> object:
        ...

    @overload
    def get(*args, **kwargs) -> Any:
        ...

    def get_raw_stream_buffer(self) -> Buffer:
        ...

    def get_stream_buffer(self, decode_level: StreamDecodeLevel = ...) -> Buffer:
        ...

    def is_owned_by(self, possible_owner: Pdf) -> bool:
        ...

    def items(self) -> Iterable[Tuple[str, Object]]:
        ...

    def keys(self) -> Set[str]:
        ...

    def page_contents_add(self, contents: Object, prepend: bool = ...) -> None:
        ...

    def page_contents_coalesce(self) -> None:
        ...

    @staticmethod
    def parse(stream: bytes, description: str = ...) -> Object:
        ...

    def read_bytes(self, decode_level: StreamDecodeLevel = ...) -> bytes:
        ...

    def read_raw_bytes(self) -> bytes:
        ...

    def same_owner_as(self, other: Object) -> bool:
        ...

    def to_json(self, dereference: bool = ...) -> bytes:
        ...

    def unparse(self, resolved: bool = ...) -> bytes:
        ...

    def wrap_in_array(self) -> Object:
        ...

    def __bytes__(self) -> bytes:
        ...

    @overload
    def __contains__(self, arg0: Object) -> bool:
        ...

    @overload
    def __contains__(self, arg0: str) -> bool:
        ...

    @overload
    def __contains__(*args, **kwargs) -> Any:
        ...

    def __copy__(self) -> Object:
        ...

    def __delattr__(self, arg0: str) -> None:
        ...

    @overload
    def __delitem__(self, arg0: str) -> None:
        ...

    @overload
    def __delitem__(self, arg0: Object) -> None:
        ...

    @overload
    def __delitem__(self, arg0: int) -> None:
        ...

    @overload
    def __delitem__(*args, **kwargs) -> Any:
        ...

    def __dir__(self) -> list:
        ...

    def __eq__(self, other: Any) -> bool:
        ...

    def __getattr__(self, arg0: str) -> Object:
        ...

    @overload
    def __getitem__(self, arg0: str) -> Object:
        ...

    @overload
    def __getitem__(self, arg0: Object) -> Object:
        ...

    @overload
    def __getitem__(self, arg0: int) -> Object:
        ...

    @overload
    def __getitem__(*args, **kwargs) -> Any:
        ...

    def __hash__(self) -> int:
        ...

    def __iter__(self) -> Iterable[Object]:
        ...

    def __len__(self) -> int:
        ...

    def __setattr__(self, arg0: str, arg1: object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: str, arg1: Object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: Object, arg1: Object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: str, arg1: object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: Object, arg1: object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: int, arg1: Object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: int, arg1: object) -> None:
        ...

    @overload
    def __setitem__(*args, **kwargs) -> Any:
        ...

    @property
    def _objgen(self) -> Tuple[int, int]:
        ...

    @property
    def _type_code(self) -> ObjectType:
        ...

    @property
    def _type_name(self) -> str:
        ...

    @property
    def images(self) -> _ObjectMapping:
        ...

    @property
    def is_indirect(self) -> bool:
        ...

    @property
    def is_rectangle(self) -> bool:
        ...

    @property
    def objgen(self) -> Tuple[int, int]:
        ...

    @property
    def stream_dict(self) -> Object:
        ...

    @stream_dict.setter
    def stream_dict(self, val: Object) -> None:
        ...


class Operator(Object):
    ...


class ObjectStreamMode(Enum):
    disable: Any = ...
    generate: Any = ...
    preserve: Any = ...


class ObjectType(Enum):
    array: Any = ...
    boolean: Any = ...
    dictionary: Any = ...
    inlineimage: Any = ...
    integer: Any = ...
    name: Any = ...
    null: Any = ...
    operator: Any = ...
    real: Any = ...
    reserved: Any = ...
    stream: Any = ...
    string: Any = ...
    uninitialized: Any = ...


class Page:
    _repr_mimebundle_: Any = ...

    def __init__(self, arg0: Object) -> None:
        ...

    def _get_cropbox(self, arg0: bool) -> Object:
        ...

    def _get_mediabox(self, arg0: bool) -> Object:
        ...

    def _get_trimbox(self, arg0: bool) -> Object:
        ...

    def add_content_token_filter(self, tf) -> None:
        ...

    def as_form_xobject(self, handle_transformations: bool = ...) -> Object:
        ...

    def contents_coalesce(self) -> None:
        ...

    def externalize_inline_images(self, min_size: int = ...) -> None:
        ...

    def get_filtered_contents(self, tf: TokenFilter) -> bytes:
        ...

    def parse_contents(self, arg0: StreamParser) -> None:
        ...

    def remove_unreferenced_resources(self) -> None:
        ...

    def rotate(self, angle: int, relative: bool) -> None:
        ...

    @property
    def _images(self) -> _ObjectMapping:
        ...

    @property
    def cropbox(self) -> Any:
        ...

    @cropbox.setter
    def cropbox(self, val: Any) -> None:
        ...

    @property
    def mediabox(self) -> Any:
        ...

    @mediabox.setter
    def mediabox(self, val: Any) -> None:
        ...

    @property
    def obj(self) -> Object:
        ...

    @property
    def trimbox(self) -> Any:
        ...

    @trimbox.setter
    def trimbox(self, val: Any) -> None:
        ...


class PageList:
    def __init__(self, *args, **kwargs) -> None:
        ...

    def append(self, page: object) -> None:
        ...

    @overload
    def extend(self, other: PageList) -> None:
        ...

    @overload
    def extend(self, iterable: Iterable[Object]) -> None:
        ...

    def insert(self, index: int, obj: object) -> None:
        ...

    def p(self, pnum: int) -> Object:
        ...

    def remove(self, **kwargs) -> None:
        ...

    def reverse(self) -> None:
        ...

    @overload
    def __delitem__(self, arg0: int) -> None:
        ...

    @overload
    def __delitem__(self, arg0: slice) -> None:
        ...

    @overload
    def __delitem__(*args, **kwargs) -> Any:
        ...

    @overload
    def __getitem__(self, arg0: int) -> Object:
        ...

    @overload
    def __getitem__(self, arg0: slice) -> list:
        ...

    @overload
    def __getitem__(*args, **kwargs) -> Any:
        ...

    def __iter__(self) -> PageList:
        ...

    def __len__(self) -> int:
        ...

    def __next__(self) -> Object:
        ...

    @overload
    def __setitem__(self, arg0: int, arg1: Object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: slice, arg1: Iterable[Object]) -> None:
        ...

    @overload
    def __setitem__(*args, **kwargs) -> Any:
        ...


class PasswordError(Exception):
    ...


class ForeignObjectError(Exception):
    ...


class Pdf:
    _attach: Any = ...
    _repr_mimebundle_: Any = ...
    add_blank_page: Any = ...
    check: Any = ...
    close: Any = ...
    make_stream: Any = ...
    open: Any = ...
    open_metadata: Any = ...
    open_outline: Any = ...
    save: Any = ...
    __enter__: Any = ...
    __exit__: Any = ...

    def __init__(self, *args, **kwargs) -> None:
        ...

    def _add_page(self, page: Object, first: bool = ...) -> None:
        ...

    def _add_page_at(self, arg0: Object, arg1: bool, arg2: Object) -> None:
        ...

    def _decode_all_streams_and_discard(self) -> None:
        ...

    def _get_object_id(self, arg0: int, arg1: int) -> Object:
        ...

    def _open(self, *args, **kwargs) -> Any:
        ...

    def _process(self, arg0: str, arg1: bytes) -> None:
        ...

    def _remove_page(self, arg0: Object) -> None:
        ...

    def _replace_object(self, arg0: int, arg1: int, arg2: Object) -> None:
        ...

    def _save(
        self,
        filename: object,
        static_id: bool = ...,
        preserve_pdfa: bool = ...,
        min_version: object = ...,
        force_version: object = ...,
        fix_metadata_version: bool = ...,
        compress_streams: bool = ...,
        stream_decode_level: object = ...,
        object_stream_mode: ObjectStreamMode = ...,
        normalize_content: bool = ...,
        linearize: bool = ...,
        qdf: bool = ...,
        progress: object = ...,
        encryption: object = ...,
        samefile_check: bool = ...,
    ) -> None:
        ...

    def _swap_objects(self, arg0: Tuple[int, int], arg1: Tuple[int, int]) -> None:
        ...

    def check_linearization(self, stream: object = ...) -> bool:
        ...

    def copy_foreign(self, h: Object) -> Object:
        ...

    @overload
    def get_object(self, objgen: Tuple[int, int]) -> Object:
        ...

    @overload
    def get_object(self, objid: int, gen: int) -> Object:
        ...

    @overload
    def get_object(*args, **kwargs) -> Any:
        ...

    def get_warnings(self) -> list:
        ...

    @overload
    def make_indirect(self, h: T) -> T:
        ...

    @overload
    def make_indirect(self, obj: object) -> Object:
        ...

    def new(self, *args, **kwargs) -> Any:
        ...

    def remove_unreferenced_resources(self) -> None:
        ...

    def show_xref_table(self) -> None:
        ...

    @property
    def Root(self) -> Object:
        ...

    @property
    def _allow_accessibility(self) -> bool:
        ...

    @property
    def _allow_extract(self) -> bool:
        ...

    @property
    def _allow_modify_all(self) -> bool:
        ...

    @property
    def _allow_modify_annotation(self) -> bool:
        ...

    @property
    def _allow_modify_assembly(self) -> bool:
        ...

    @property
    def _allow_modify_form(self) -> bool:
        ...

    @property
    def _allow_modify_other(self) -> bool:
        ...

    @property
    def _allow_print_highres(self) -> bool:
        ...

    @property
    def _allow_print_lowres(self) -> bool:
        ...

    @property
    def _encryption_data(self) -> dict:
        ...

    @property
    def _pages(self) -> Any:
        ...

    @property
    def allow(self) -> Any:
        ...

    @property
    def docinfo(self) -> Object:
        ...

    @docinfo.setter
    def docinfo(self, val: Object) -> None:
        ...

    @property
    def encryption(self) -> Any:
        ...

    @property
    def extension_level(self) -> int:
        ...

    @property
    def filename(self) -> str:
        ...

    @property
    def is_encrypted(self) -> bool:
        ...

    @property
    def is_linearized(self) -> bool:
        ...

    @property
    def objects(self) -> Any:
        ...

    @property
    def pages(self) -> PageList:
        ...

    @property
    def pdf_version(self) -> str:
        ...

    @property
    def root(self) -> Object:
        ...

    @property
    def trailer(self) -> Object:
        ...


class PdfError(Exception):
    ...


class StreamDecodeLevel(Enum):
    all: Any = ...
    generalized: Any = ...
    none: Any = ...
    specialized: Any = ...


class StreamParser:
    def __init__(self) -> None:
        ...

    def handle_eof(self) -> None:
        ...

    def handle_object(self, arg0: Object) -> None:
        ...


class Token:
    def __init__(self, arg0: TokenType, arg1: bytes) -> None:
        ...

    def __eq__(self, other: Any) -> bool:
        ...

    @property
    def error_msg(self) -> str:
        ...

    @property
    def raw_value(self) -> bytes:
        ...

    @property
    def type_(self) -> TokenType:
        ...

    @property
    def value(self) -> str:
        ...


class TokenFilter(_QPDFTokenFilter):
    def __init__(self) -> None:
        ...

    def handle_token(self, token: Token = ...) -> object:
        ...


class TokenType(Enum):
    array_close: Any = ...
    array_open: Any = ...
    bad: Any = ...
    bool: Any = ...
    brace_close: Any = ...
    brace_open: Any = ...
    comment: Any = ...
    dict_close: Any = ...
    dict_open: Any = ...
    eof: Any = ...
    inline_image: Any = ...
    integer: Any = ...
    name: Any = ...
    null: Any = ...
    real: Any = ...
    space: Any = ...
    string: Any = ...
    word: Any = ...


class _ObjectList:
    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, arg0: _ObjectList) -> None:
        ...

    @overload
    def __init__(self, arg0: Iterable) -> None:
        ...

    @overload
    def __init__(*args, **kwargs) -> None:
        ...

    def append(self, x: Object) -> None:
        ...

    def clear(self) -> None:
        ...

    def count(self, x: Object) -> int:
        ...

    @overload
    def extend(self, L: _ObjectList) -> None:
        ...

    @overload
    def extend(self, L: Iterable[Object]) -> None:
        ...

    def insert(self, i: int, x: Object) -> None:
        ...

    @overload
    def pop(self) -> Object:
        ...

    @overload
    def pop(self, i: int) -> Object:
        ...

    @overload
    def pop(*args, **kwargs) -> Any:
        ...

    def remove(self, x: Object) -> None:
        ...

    def __bool__(self) -> bool:
        ...

    def __contains__(self, x: Object) -> bool:
        ...

    @overload
    def __delitem__(self, arg0: int) -> None:
        ...

    @overload
    def __delitem__(self, arg0: slice) -> None:
        ...

    @overload
    def __delitem__(*args, **kwargs) -> Any:
        ...

    def __eq__(self, other: Any) -> bool:
        ...

    @overload
    def __getitem__(self, s: slice) -> _ObjectList:
        ...

    @overload
    def __getitem__(self, arg0: int) -> Object:
        ...

    @overload
    def __getitem__(*args, **kwargs) -> Any:
        ...

    def __iter__(self) -> Iterator[Object]:
        ...

    def __len__(self) -> int:
        ...

    def __ne__(self, other: Any) -> bool:
        ...

    @overload
    def __setitem__(self, arg0: int, arg1: Object) -> None:
        ...

    @overload
    def __setitem__(self, arg0: slice, arg1: _ObjectList) -> None:
        ...

    @overload
    def __setitem__(*args, **kwargs) -> Any:
        ...


class _ObjectMapping:
    get: Any = ...
    keys: Any = ...
    values: Any = ...
    __contains__: Any = ...

    def __init__(self) -> None:
        ...

    def items(self) -> Iterator:
        ...

    def __bool__(self) -> bool:
        ...

    def __delitem__(self, arg0: str) -> None:
        ...

    def __getitem__(self, arg0: str) -> Object:
        ...

    def __iter__(self) -> Iterator:
        ...

    def __len__(self) -> int:
        ...

    def __setitem__(self, arg0: str, arg1: Object) -> None:
        ...


class _QPDFTokenFilter:
    def __init__(self, *args, **kwargs) -> None:
        ...
